// Copyright (c) 2002 Tom Barry.  All rights reserved.
//      trbarry@trbarry.com
// Requires Avisynth source code to compile for Avisynth
// Avisynth Copyright 2000 Ben Rudiak-Gould.
//      http://www.math.berkeley.edu/~benrg/avisynth.html
/////////////////////////////////////////////////////////////////////////////
//
//	This file is subject to the terms of the GNU General Public License as
//	published by the Free Software Foundation.  A copy of this license is
//	included with this software distribution in the file COPYING.  If you
//	do not have a copy, you may obtain a copy by writing to the Free
//	Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
//	This software is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details
//  
//  Also, this program is "Philanthropy-Ware".  That is, if you like it and 
//  feel the need to reward or inspire the author then please feel free (but
//  not obligated) to consider joining or donating to the Electronic Frontier
//  Foundation. This will help keep cyber space free of barbed wire and bullsh*t.  
//
/////////////////////////////////////////////////////////////////////////////
// Change Log
//
// Date          Developer             Changes
//
// 02 Aug 2002   Tom Barry		       Create STMedianFilter V 0.0.1.0
//
/////////////////////////////////////////////////////////////////////////////
// CVS Log
//
//
/////////////////////////////////////////////////////////////////////////////
#include "avisynth.h"
#include <malloc.h>
#include <string.h>

#define BOOL unsigned int
#define BYTE unsigned char

const __int64 YMask = 0x00ff00ff00ff00ff;
const __int64 UVMask = 0xff00ff00ff00ff00;

class STMedianFilter : public GenericVideoFilter {
    int  Max_Frames;
    __int64  SThreshHoldF;
    __int64  SThreshHoldE;
    __int64  TThreshHoldF;
    __int64  TThreshHoldM;
    __int64  HThreshHoldF;
    __int64  HThreshHoldE;
	int  useFrame;

    PVideoFrame src;
    PVideoFrame dst;
    PVideoFrame prevSrc;
    PVideoFrame nextSrc;
    unsigned char* dstp;


    int src_pitch;
    int dst_pitch;
    int rowsize;
    int RowPixels;
    int height;
    int FldHeight;
	bool SSE2enabled;
	bool SSEMMXenabled;
	bool _3DNOWenabled;
	int __stdcall Fieldcopy(void *dest, const void *src, size_t count, 
				int rows, int dst_pitch, int src_pitch);
	BOOL GoFilterIt(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp, const BYTE* pPrev, const BYTE* pNext,
			int FldHeight);					// go H expand lines into even output lines

public:
    STMedianFilter(PClip _child, 
		int _SThreshHoldF, int _SThreshHoldE,
		int _TThreshHoldF, int _TThreshHoldM,
		int _HThreshHoldF, int _HThreshHoldE,
		 IScriptEnvironment* env) : GenericVideoFilter(_child)
    {
		_int64 i;
	
		i = _SThreshHoldF;	
        SThreshHoldF = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 
		i = _SThreshHoldE;	
        SThreshHoldE = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 
		i = _TThreshHoldF;	
        TThreshHoldF = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 
		i = _TThreshHoldM;	
        TThreshHoldM = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 
		i = _HThreshHoldF;	
        HThreshHoldF = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 
		i = _HThreshHoldE;	
        HThreshHoldE = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 
        Max_Frames = vi.num_frames;
		SSEMMXenabled = (env->GetCPUFlags() & CPUF_INTEGER_SSE) !=0;


	    if (!vi.IsYUY2())
		{
			env->ThrowError("STMedianFilter: Supports YUY2 color format only");
		}
	    if (!SSEMMXenabled)
		{
			env->ThrowError("STMedianFilter: Only SSEMMX CPU's currently supported");
		}

    }

    PVideoFrame __stdcall GetFrame(int inFrame, IScriptEnvironment* env);

};


PVideoFrame __stdcall STMedianFilter::GetFrame(int inFrame, IScriptEnvironment* env) 
{
PVideoFrame PrevSrc;
PVideoFrame NextSrc;

const BYTE* srcp;
const BYTE* pPrev;
const BYTE* pNext;
	
	useFrame = inFrame;

    src = child->GetFrame(useFrame, env);
	srcp = src->GetReadPtr();

	if (useFrame < 1 || useFrame >= Max_Frames-1)
	{
		return src;						// don't filter first and last frames
	}
/*	
	prevSrc = child->GetFrame(useFrame-1, env);
		pWeaveSrcP = TopFirst
			? prevSrc->GetReadPtr() + src_pitch
*/
	PrevSrc = child->GetFrame(useFrame-1, env);
	pPrev = PrevSrc->GetReadPtr();
	NextSrc = child->GetFrame(useFrame+1, env);
	pNext = NextSrc->GetReadPtr();
    dst = env->NewVideoFrame(vi);
    dstp = dst->GetWritePtr();
    dstp = dst->GetWritePtr();
    src_pitch = src->GetPitch();
    dst_pitch = dst->GetPitch();
    rowsize = dst->GetRowSize();
	RowPixels = rowsize >> 1;
    height = dst->GetHeight();

	GoFilterIt(src_pitch, dst_pitch,
		rowsize, srcp, dstp, pPrev, pNext,
		height);					// go H expand lines into even output lines

/*	
	Fieldcopy(pWeaveDest + (height-2)*dst_pitch,			// copy bottom line	
		pCopySrc+(FldHeight-1)*src_pitch, rowsize, 
				1, dst_pitch, src_pitch);
*/
	
	__asm emms;

    return dst;
}

#define IS_SSE
#define SSE_TYPE SSE


#define ChkBetterNeighbors(PADDR1, PADDR2) \
__asm { \
		__asm movq	mm0, qword ptr[(PADDR1)] /* our 4 pixels */ \
		__asm movq mm1, qword ptr[(PADDR2)] /* our pixel2 value */ \
		__asm movq mm2, mm0			/* another copy of our pixel1 value */ \
		__asm pminub mm0, mm1       /* minimum value of the 2 */ \
		__asm pmaxub mm1, mm2       /* maximum value of the 2 */ \
		__asm movq mm2, mm1			/* another copy of our max value */ \
		__asm psubusb mm2, mm0 	    /* the diff */ \
		__asm movq mm3, mm2			/* another copy of our our weights (diff) */ \
		__asm psubusb mm3, mm5		/* nonzero where old weights lower, else 0 */ \
		__asm pxor mm4, mm4			/* zero */ \
		__asm pcmpeqb mm3, mm4		/* now ff where new better, else 00	*/ \
		__asm pcmpeqb mm4, mm3		/* here ff where old better, else 00 */ \
		__asm pand	mm0, mm3		/* keep only better min pixels */ \
		__asm pand	mm1, mm3		/* keep only better max pixels */ \
		__asm pand  mm2, mm3		/* and weights */ \
		__asm pand	mm5, mm4		/* keep only better old weights */ \
		__asm pand  mm6, mm4        /* keep only better old min  */ \
		__asm pand  mm7, mm4        /* keep only better old max  */ \
		__asm por	mm5, mm2		/* merge new & old weights */ \
		__asm por	mm6, mm0        /* merge new & old min */\
		__asm por	mm7, mm1		/* merge new & old max */ \
}




BOOL	STMedianFilter::GoFilterIt(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp, const BYTE* _pPrev, const BYTE* _pNext,
			int FldHeight)					// go H expand lines into even output lines
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-10) >> 3);
__int64 SThreshHoldFW = SThreshHoldF;
__int64 SThreshHoldEW = SThreshHoldE;
__int64 TThreshHoldFW = TThreshHoldF;
__int64 TThreshHoldMW = TThreshHoldM;
__int64 HThreshHoldFW = HThreshHoldF;
__int64 HThreshHoldEW = HThreshHoldE;
__int64 SaveTemp;
		
	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pit*(FldHeight-1),
		pSrc+src_pit*(FldHeight-1), rowsize);  // and last

	pDest += dst_pit;


	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pit 
		mov		edx, pDest 

		// do first and last qwords, just move
		movq	mm0, qword ptr[ebx+ecx]		// just move 1st 4 pixels
		movq	qword ptr[edx], mm0
		
		mov		eax, rowsize
		add		ebx, eax
		movq	mm0, qword ptr[ebx+ecx-8]		// just move last 4 pixels
		movq	qword ptr[edx+eax-8], mm0

		sub		ebx, eax						// restore
		
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	
		pcmpeqb	mm5, mm5		// init diff 255's
		pxor	mm6, mm6		// init min to 0
		pcmpeqb	mm7, mm7		// init max 255's

// Search between prev and next frame - Temporal
        ChkBetterNeighbors(edi-2, esi+2*ecx+2)  // up left, down right
        ChkBetterNeighbors(edi+2, esi+2*ecx-2)  // up right, down left
        ChkBetterNeighbors(edi+2*ecx-2, esi+2)  // down left, up right
        ChkBetterNeighbors(edi+2*ecx+2, esi-2)   // down right, up left   
        ChkBetterNeighbors(edi+ecx-2, esi+ecx+2) // left, right
        ChkBetterNeighbors(edi+ecx+2, esi+ecx-2) // right, left

		por		mm5, UVMask						// invalidate non-vertical chroma matches
		ChkBetterNeighbors(edi+ecx, esi+ecx)	// center, in old and new
        ChkBetterNeighbors(edi+2*ecx, esi)		// down, up
        ChkBetterNeighbors(edi, esi+2*ecx)		// up, down

// We now know the min and max of the closest together pair of neighbors
// We will clip our current 4 pixels to (each) be withing this range if it would
// not cause them to change by more than the ThreshHold user parm
		movq	mm0, qword ptr[ebx+ecx]		// our curr pixels vals
		movq	mm1, mm0
//#define ALSO_AVG							// don't know if I also want this
#ifdef ALSO_AVG
		movq	mm3, mm6
		pavgb	mm3, mm7					// real median of sorounding pixels
		pavgb   mm1, mm3					// see if averaging this in helps
#endif

		pminub	mm1, mm7					// make it no larger than max neighbor
		pmaxub	mm1, mm6					// and now smaller than min neighbor
		movq	mm3, mm1
		movq	mm4, mm0
		psubusb	mm3, mm0					// changed - orig
		psubusb mm4, mm1					// orig - changed
		por		mm3, mm4					// the abs diff
		psubusb	mm3, TThreshHoldFW			// did it change too much, Y where nonzero
		psubusb mm5, TThreshHoldMW			// Too much motion? Y where nonsero
		por		mm3, mm4					// Don't clip where either nonzero
		pxor mm4, mm4						// zero
		pcmpeqb mm3, mm4					// now ff where new value should be used, else 00	
//		pand	mm3, YMask					// but only filter Luma (now allow chroma too)
		pcmpeqb mm4, mm3					// here ff where old better, else 00 
		pand	mm1, mm3					// use new value for these pixels
		pand	mm0, mm4					// use old value for these pixels
		por		mm0, mm1					// combine them
		movq	SaveTemp, mm0				// save results for later

		// Do again to search spacial (except horizontal, both pixels in curr frame
		pcmpeqb	mm5, mm5		// init diff 255's
		pxor	mm6, mm6		// init min to 0
		pcmpeqb	mm7, mm7		// init max 255's

        ChkBetterNeighbors(ebx-2, ebx+2*ecx+2)  // up left, down right
        ChkBetterNeighbors(ebx+2, ebx+2*ecx-2)  // up right, down left

		por		mm5, UVMask						// invalidate non-vertical chroma matches
        ChkBetterNeighbors(ebx, ebx+2*ecx)	// up, down

		movq	mm0, SaveTemp				// our curr pixels vals
		movq	mm1, mm0
#define ALSO_AVG
#ifdef ALSO_AVG
		movq	mm3, mm6
		pavgb	mm3, mm7					// real median of sorounding pixels
		pavgb   mm1, mm3					// see if averaging this in helps
#endif

		pminub	mm1, mm7					// make it no larger than max neighbor
		pmaxub	mm1, mm6					// and now smaller than min neighbor

		movq	mm3, mm1
		movq	mm4, mm0
		psubusb	mm3, mm0					// changed - orig
		psubusb mm4, mm1					// orig - changed
		por		mm3, mm4					// the abs diff
		psubusb	mm3, SThreshHoldFW			// did it change too much, Y where nonzero
		psubusb mm5, SThreshHoldEW			// Too much motion? Y where nonsero
		por		mm3, mm4					// Don't clip where either nonzero
		pxor	mm4, mm4					// zero
		pcmpeqb mm3, mm4					// now ff where new value should be used, else 00	
//		pand	mm3, YMask					// but only filter Luma (now allow chroma too)
		pcmpeqb mm4, mm3					// here ff where old better, else 00 
		pand	mm1, mm3					// use new value for these pixels
		pand	mm0, mm4					// use old value for these pixels
		por		mm0, mm1					// combine them, leave answer sitting in mm0


//      ChkBetterNeighbors(ebx+ecx-2, ebx+ecx+2) // left, right - but do without macro
		movq	mm6, qword ptr[ebx+ecx-2]	// our 4 pixels  
		movq	mm7, qword ptr[ebx+ecx+2]		// our pixel2 value  
		movq	mm2, mm6						// another copy of our pixel1 value 
		pminub	mm6, mm7						// minimum value of the 2 
		pmaxub	mm7, mm2						// maximum value of the 2  
		movq	mm5, mm7						// another copy of our max value  
		psubusb mm5, mm6 					// the diff  

		movq	mm1, mm0
#ifdef ALSO_AVG
		movq	mm3, mm6
		pavgb	mm3, mm7					// real median of sorounding pixels
		pavgb   mm1, mm3					// see if averaging this in helps
#endif

		pminub	mm1, mm7					// make it no larger than max neighbor
		pmaxub	mm1, mm6					// and no smaller than min neighbor

		movq	mm3, mm1
		movq	mm4, mm0
		psubusb	mm3, mm0					// changed - orig
		psubusb mm4, mm1					// orig - changed
		por		mm3, mm4					// the abs diff
		psubusb	mm3, HThreshHoldFW			// did it change too much, Y where nonzero
		psubusb mm5, HThreshHoldEW			// Too much motion? Y where nonsero
		por		mm3, mm4					// Don't clip where either nonzero
		pxor mm4, mm4						// zero
		pcmpeqb mm3, mm4					// now ff where new value should be used, else 00	
		pand	mm3, YMask					// but only filter Luma here
		pcmpeqb mm4, mm3					// here ff where old better, else 00 
		pand	mm1, mm3					// use new value for these pixels
		pand	mm0, mm4					// use old value for these pixels
		por		mm0, mm1					// combine them


		movntq	qword ptr[edx], mm0			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ
		}
	// adjust for next line
	pSrc  += src_pit;
	pPrev  += src_pit;
	pNext  += src_pit;
	pDest += dst_pit;
	}
	
	return 0;


}

int STMedianFilter::Fieldcopy(void *dest, const void *src, size_t count, 
				int rows, int dst_pitch, int src_pitch)
{
BYTE* pDest = (BYTE*) dest;
BYTE* pSrc = (BYTE*) src;
	
	for (int i=0; i < rows; i++)
	{
		memcpy(pDest, pSrc, count);
		pSrc += src_pitch;
		pDest += dst_pitch;
	}
	return 0;
}


AVSValue __cdecl Create_STMedianFilter(AVSValue args, void* user_data,
IScriptEnvironment* env) 
{
    return new STMedianFilter(args[0].AsClip(), args[1].AsInt(), args[2].AsInt(),
		args[3].AsInt(), args[4].AsInt(), args[5].AsInt(), args[6].AsInt(),env);
}


extern "C" __declspec(dllexport) const char* __stdcall
AvisynthPluginInit(IScriptEnvironment* env) {
    env->AddFunction("STMedianFilter", "ciiiiii", Create_STMedianFilter, 0);
    return "`STMedianFilter' Clip in Space and Time";
}

